
// This class handles the timing and playback of animation frames
// It is threaded, so it will continue to move the frames forward
// while the user does other things.

import java.io.*;
import java.awt.*;
import java.util.*;
import java.awt.image.*;
import com.sun.java.swing.*;
import com.sun.java.swing.JPanel;

public class AnimationPlayback extends Thread 
{   
    public sphclient myparent;
    
    private int delay = 500;    // delay between frames, in milliseconds
    private boolean playing = false;
    
    public AnimationPlayback(Frame parent)
    {
        myparent = (sphclient) parent;       
    }
    
    public void run() 
    {
        while (true)
        {
            try {               
                Thread.sleep(delay);
            }
            catch (Exception e)
            {
                System.out.println("Error trying to sleep!");
            }
            
            int current_frame = myparent.frame_scroll.getValue();
            
            if (playing)
            {
                current_frame++;
                
                // see if we have reached the end of the frames
                if (current_frame > myparent.frame_scroll.getMaximum())
                {    
                    current_frame--;
                    playing = false;
                }
                else
                {
                    myparent.frame_scroll.setValue(current_frame);
                    myparent.update_canvases();
                }
            }
        }
    }
    
    public void playit()
    {
        playing = true;
    }
    
    public void stopit()
    {
        playing = false;       
    }
    
    public void rewindit()
    {
        playing = false;
        myparent.frame_scroll.setValue(0);
    }

    public int getDelay()
    {
        return delay;
    }
    
    public void setDelay(int d)
    {
        delay = d;
    }
}