#include "Box.h"

//******************************************************************
//******************************************************************
Box::Box() : Shape(BOX)
{}

//******************************************************************
//******************************************************************
Box::Box(Point b, double w, double h) : Shape(BOX)
{
	lowerLeftCorner = b;
				
	width = w;
	height = h;
}

//******************************************************************
//******************************************************************
Box::Box(const Box & L)
{
	lowerLeftCorner = L.lowerLeftCorner;
	width = L.getWidth(); 
	height = L.getHeight();
}



//******************************************************************
//******************************************************************
Box & Box::operator = (const Box &L)
{
	int i;

	if(this == &L)
		return *this;
	
	width = L.getWidth();
	height = L.getHeight();
	type = L.getType(); 
	lowerLeftCorner = L.getCornerPoint(); 
	
	return (*this);
}

//******************************************************************
//******************************************************************
ostream & operator << ( ostream & str, const Box & L)
{
	str << "Lowerleft = " << L.getCornerPoint() << "  height = "
		<< L.getHeight() << "  width = " << L.getWidth() << endl;

	return str;
}	

//******************************************************************
//******************************************************************
ofstream & operator << ( ofstream & fStr, const Box & L)
{
	fStr << "Lowerleft = " << L.getCornerPoint() << "  height = "
		<< L.getHeight() << "  width = " << L.getWidth() << endl;

	return fStr;
}	

//******************************************************************
//******************************************************************
double Box::area() const
{
	return (height * width);
}

//******************************************************************
// Checks to see if the point (x, y) is inside the region of space
// occupied by this shape.
//******************************************************************
int Box::isInside(double x, double y)
{
	if ( (x <= lowerLeftCorner.getX() + width )  &&
	     (x >= lowerLeftCorner.getX()         )  &&
	     (y <= lowerLeftCorner.getY() + height)  &&
	     (y >= lowerLeftCorner.getY()         ) )
		return 1;
	else
		return 0;
}


//******************************************************************
// Used for interaction between free and bound Smooth Particles. 
//******************************************************************
double Box::calcRatioOfDistances(Point aPos, Point bPos)
{
	double dist1, dist2;

	dist1 = fabs(aPos.getY() - lowerLeftCorner.getY());
	dist2 = fabs(aPos.getY() - lowerLeftCorner.getY() + height);
	
	if (dist1 < dist2)
		return fabs((bPos.getY() - lowerLeftCorner.getY()) / dist1);	
	else
		return fabs((bPos.getY() - lowerLeftCorner.getY()) / dist2);	
}
