//***********************************************************
//
// A Cell contains two linked lists; one for the free particles
// and one for the bound particles. 
//
// Written by: Brian Schlatter
// Last updated: 6/12/1999
//
//********************************************************** 

#ifndef CELL_HPP
#define CELL_HPP

#include "FastLinkedList.h"


class Cell {
	public:
		Cell();
		~Cell();

		//Get the size of each list.
		int getNumFreePts() { return Free.length(); }
		int getNumBndPts()  { return Bnd.length();  }

		//Methods to move pointers along each list.
		void cur2NextFreeItem()     { Free.nextCur();   }
		void cur2NextBndItem()      { Bnd.nextCur();    }
		void other2NextFreeItem()   { Free.nextOther(); }
		void other2NextBndItem()    { Bnd.nextOther();  }
		void cur2FirstFreeItem()    { Free.firstCur();  }
		void cur2FirstBndItem()     { Bnd.firstCur();   }
		void other2FirstFreeItem()  { Free.firstOther();}
		void other2FirstBndItem()   { Bnd.firstOther(); }
				
		//Add NewItem to the front of the respective list.
		void pushFree(itemType item) { Free.push(item); }
		void pushBnd(itemType item)  { Bnd.push(item);  }

		//Remove the item that "Cur" is pointing to from the list.
		void removeCurrentFreeItem() { Free.deleteCur();   }
		void removeOtherFreeItem()   { Free.deleteOther(); }
		void removeCurrentBndItem()  { Bnd.deleteCur();    }

		// get rid of this cell's free linked list
		void deleteFree() { Free.deleteAll(); }
		void deleteBnd() { Bnd.deleteAll(); }
	
		//Get the item that pointer Cur/Other is pointing to in
		//the respective list.
		itemType getCurrentFreeItem() { return Free.getCur();   }
		itemType getCurrentBndItem()  { return Bnd.getCur();    }
		itemType getOtherFreeItem()   { return Free.getOther(); }
		itemType getOtherBndItem()    { return Bnd.getOther();  }

		void print() { Free.print(); Bnd.print(); }	
		
	private:
		FastLinkedList Free;
		FastLinkedList Bnd;


};

#endif
