#include "FastLinkedList.h"


//*****************************************************************
//* What each node looks like: Two pointers and an itemType value.
//*****************************************************************
struct listNode
{
	itemType Item;
	ptrType      Next;  // Needed for normal list traversal.
	ptrType      Prev;  // Needed for O(1) item removal.
};


//****************************************************************
//* Constructor: Initialize everything to 0 or NULL.
//*****************************************************************
FastLinkedList::FastLinkedList(): Size(0),  Head(NULL), Tail(NULL), 
Cur(NULL), Other(NULL){ }


//*****************************************************
//* Deletes all nodes.
//*****************************************************
void FastLinkedList::deleteAll()
{
	//Delete the list
	if(Size > 0)
	{
		Cur = Head;
		while(Size != 0)
			deleteCur();
	}

	Head = NULL;	
	Tail = NULL;
	Cur = NULL;
	Other = NULL;
}

//*****************************************************
//* Deletes all nodes.
//*****************************************************
FastLinkedList::~FastLinkedList()
{
        deleteAll();
}

//***********************************************************
//* The following methods move the respective pointer to the next
//* item.
//***********************************************************
void FastLinkedList::nextCur() 
{ if(Cur != NULL) Cur = Cur->Next; }


void FastLinkedList::nextOther() 
{ if(Other != NULL) Other = Other->Next; }




//************************************************************
//*The following methods move the respective pointer to the front 
//* of the list.
//************************************************************
void FastLinkedList::firstCur()
{ Cur = Head; }

void FastLinkedList::firstOther()
{ Other = Head; }



//*************************************************************
//* The following method adds NewItem to the end of the 
//* linked list.
//*************************************************************
void FastLinkedList::push(itemType NewItem)
{
	Size++;
	
	// Add to an empty list.
	if(Tail == NULL)
	{	
		ptrType First = new listNode;
		assert(First != NULL);
		First->Item = NewItem;
		Head = First;
		Tail = First;
		Cur = First;
		Other = First;
		First->Next = NULL;
		First->Prev = NULL;
	}
	// Add to a non-empty list.
	else
	{
		ptrType New = new listNode;
		assert(New != NULL);
		New->Item = NewItem;
		Tail->Next = New;
		New->Prev = Tail;
		New->Next = NULL;
		Tail = Tail->Next;
	}	

}




//***************************************************************
//* The following method deletes the node at which "Cur" points.
//**************************************************************
void FastLinkedList::deleteCur()
{
	if(Cur != NULL)
	{	Size--;
		
		// Check for the "first" node.
		if(Cur->Prev != NULL)
			Cur->Prev->Next = Cur->Next;
		// Check for the "last" node.
		if(Cur->Next != NULL)
			Cur->Next->Prev = Cur->Prev;
		
		// Now move all pointers to next node if they are
		// pointing to the 'to be' deleted node.
		if(Other == Cur)
			Other = Other->Next;
		ptrType Temp = Cur;
		Cur = Cur->Next;
		
		// A special case is if we are deleteing the first node.
		if(Temp == Head)
			Head = Head->Next;
		// The last node is also a special case.
		if (Temp == Tail)
			Tail = Tail->Prev;
		
		// Finally, free up the memory.
		Temp->Next = NULL;
		Temp->Prev = NULL;
		delete Temp;
		Temp = NULL;
	}
	
}

//***************************************************************
//* The following method deletes the node at which "Other" points.
//**************************************************************
void FastLinkedList::deleteOther()
{
	if(Other != NULL)
	{	Size--;
		// Check for the "first" node.
		if(Other->Prev != NULL)
			Other->Prev->Next = Other->Next;
		// Check for the "last" node.
		if(Other->Next != NULL)
			Other->Next->Prev = Other->Prev;
		
		// Now move all pointers to next node if they are
		// pointing to the 'to be' deleted node.
		if(Other == Cur)
			Cur = Cur->Next;

		ptrType Temp = Other;
		Other = Other->Next;

		// A special case is if we are deleteing the first node.
		if(Temp == Head)
			Head = Head->Next;
		// The last node is also a special case.
		if (Temp == Tail)
			Tail = Tail->Prev;
		
		// Finally, free up the memory.
		Temp->Next = NULL;
		Temp->Prev = NULL;
		delete Temp;
		Temp = NULL;
	}
	
}



//*********************************************************
//* The following method returns the value that is stored
//* in the node that the respective pointer is pointing to.
//*********************************************************
itemType FastLinkedList::getCur()
{
	return (Cur != NULL) ? Cur->Item : (itemType)(-1);
}


itemType FastLinkedList::getOther()
{
	return (Other != NULL) ? Other->Item : (itemType)(-1);
}


//*********************************************************
//* print out the contents of the linked list.
//*********************************************************
void FastLinkedList::print()
{
	int i;
	
	ptrType temp = Head;
	
	cout << "Head";
	for(i = 0; i < Size; i++)
	{
		cout <<  " -> " << temp->Item;
		temp = temp->Next;
	}

	cout << endl;
}

