/*
 * @(#)GifImage.java	1.1	96/03/16  
 *
 * Copyright (c) 1996 Tadashi Hamano All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. 
 *
 * I SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS
 * A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE
 * OR ITS DRIVATIVES.
 *
 *	History :
 *		1.0	96/03/14		initial version
 *		1.1	96/03/16		fix a bug (transparent)
 */

// This is based on "GifImage" code on alpha3 API.
// Listed below is the original Copyright notice.
/*
 * @(#)GifImage.java	1.24 95/05/11 Patrick Naughton, Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */
// End of the original Copyright notice.

/*-
 *	Reads GIF image from memory, construct 
 *	a MemoryImageSource object.
 *
 * The algorithm is copyright of CompuServe.
 */

import	java.awt.image.*;
import java.io.*;

/**
 * Gif Image converter
 * 
 * @version 1.1 16 Mar 1996
 * @author Tadashi Hamano
 */

public
class GifImage {

	private static final int IMAGESEP 		= 0x2c;
	private static final int EXBLOCK 		= 0x21;
	private static final int EX_GRAPHICS_CONTROL	= 0xf9;
	private static final int EX_COMMENT 		= 0xfe;
	private	static final int EX_APPLICATION 	= 0xff;
	private static final int TERMINATOR 		= 0x3b;
	private	static final int INTERLACEMASK 		= 0x40;
	private static final int COLORMAPMASK 		= 0x80;

	private int			width;
	private int			height;
	private byte			raster[];
	private IndexColorModel		icm;
	private MemoryImageSource	mis;
	private int			bPtr;
	private int			num_colors;
	private int			trans_index = -1;
	private byte			red[];
	private byte			green[];
	private byte			blue[];
	private byte			buff[];

	/**
	 * Constructor, read image from memory.
	 * modified by kdv to read from a file on disk
	 */
	public GifImage( String filename) throws GifImageException {

        int file_length = (int) (new File(filename)).length();
        
        byte src[] = new byte[1];   // instantiate just to shut up the compiler
        
        try {
            BufferedInputStream input = new BufferedInputStream(
                                            new FileInputStream(filename));       
            src = new byte[file_length];
        
            input.read(src, 0, file_length);
            
            input.close();
        }
        catch (Exception e)
        {
            System.out.println("Error reading gif file into array!");
        }
        
        
	    buff = src;
	    bPtr = 0;

	    readHeader();

	    while (true) {
		int code;
		switch (code = readBuff()) {
		  case EXBLOCK:
		    switch (code = readBuff()) {
		      case EX_GRAPHICS_CONTROL: {
			byte buf[] = new byte[6];
			if (readBuff( buf, 6) == -1) {
			    throw new GifImageException( "corrupt GIF file");
			}
			if ((buf[0] != 4) || (buf[5] != 0)) {
			    throw new GifImageException( 
				"corrupt GIF file (GCE size)");
			}
			// Get the index of the transparent color
			trans_index = buf[4] & 0xFF;
			break;
		      }
		      case EX_COMMENT:
		      case EX_APPLICATION:
		      default:
			while (true) {
			    int n = readBuff();
			    if (n == 0) {
				break;
			    }
			    byte buf[] = new byte[n];
			    if ( readBuff( buf, n) == -1) {
				throw new GifImageException( 
				    "corrupt GIF file");
			    }
			}
			break;
		    }
		    break;
		  case IMAGESEP:
		    try {
			readImage();
			icm = new IndexColorModel( 8, num_colors, 
			    red, green, blue, trans_index);
			mis = new MemoryImageSource( width, height, icm,
			    raster, 0, width);
		    } catch (ArrayIndexOutOfBoundsException e) {
			e.printStackTrace();
			throw new GifImageException( "corrupt gif file");
		    }
		    break;
		  case TERMINATOR:
		    return;
		  case -1:
		    return;
		  default:
		    throw new GifImageException( 
			"corrupt GIF file (parse) [" + code + "].");
		}
	    }
	}

	/**
	 * Returns MemoryImageSource to createImage();.
	 */
	public MemoryImageSource get() {

	    return( mis);
	}

	/**
	 * Read Image header
	 */
	private void readHeader() throws GifImageException {

	    byte buf[] = new byte[256*3];

	    if ( readBuff( buf, 13) == -1) {
		throw new GifImageException( "corrupt GIF file (header)");
	    }

	    if (( buf[0] != 'G') || ( buf[1] != 'I') || ( buf[2] != 'F')) {
		throw new GifImageException( "not a GIF file.");
	    }

	    // colormap info
	    int ch = buf[10] & 0xFF;
	    if (( ch & COLORMAPMASK) == 0) {
		throw new GifImageException( "no global colormap.");
	    }
	    num_colors = 1 << ((ch & 0x7) + 1);

	    // supposed to be NULL
	    if ( buf[12] != 0) {
		throw new GifImageException( "corrupt GIF file (nonull).");
	    }

	    // Read colors
	    if ( readBuff( buf, num_colors * 3) == -1) {
		throw new GifImageException( "corrupt Gif file (no color)");
	    }

	    red = new byte[num_colors];
	    green = new byte[num_colors];
	    blue = new byte[num_colors];

	    // Copy colors
	    for ( int i = 0; i < num_colors; i++) {
		red[i]   = buf[i*3 + 0];
		green[i] = buf[i*3 + 1];
		blue[i]  = buf[i*3 + 2];
	    }
	}

	private void readImage() throws GifImageException {

	    byte block[] = new byte[256 + 3];

	    // Read the image descriptor
	    if ( readBuff( block, 10) == -1) {
		throw new GifImageException( 
		    "corrupt GIF image.(no image descriptor)");
	    }

	    width = ( block[4] & 0xFF) | ( block[5] << 8);
	    height = ( block[6] & 0xFF) | ( block[7] << 8);
	    boolean interlace = ( block[8] & INTERLACEMASK) != 0;
	    int initCodeSize = block[9] & 0xFF;

	    byte ras[] = raster = new byte[ width * height];

	// Patrick Naughton:
	// Note that I ignore the possible existence of a local color map.
	// I'm told there aren't many files around that use them, and the
	// spec says it's defined for future use.  This could lead to an
	// error reading some files.
     	//
	// Start reading the image data. First we get the intial code size
	// and compute decompressor constant values, based on this code
	// size.
	//
	// The GIF spec has it that the code size is the code size used to
	// compute the above values is the code size given in the file,
	// but the code size used in compression/decompression is the code
	// size given in the file plus one. (thus the ++).

	// Arthur van Hoff:
	// The following narly code reads LZW compressed data blocks and
	// dumps it into the image data. The input stream is broken up into
	// blocks of 1-255 characters, each preceded by a length byte.
	// 3-12 bit codes are read from these blocks. The codes correspond to
	// entry is the hashtable (the prefix, suffix stuff), and 
	// the appropriate pixels are written to the image.

	    int clearCode = (1 << initCodeSize);
	    int eofCode = clearCode + 1;
	    int bitMask = num_colors - 1;

	    // Variables used to form reading data
	    boolean blockEnd = false;
	    int remain = 0;
	    int byteoff = 0;
	    int accumbits = 0;
	    int accumdata = 0;
	    
	    // Variables used to decompress the data
	    int codeSize = initCodeSize + 1;
	    int maxCode = 1 << codeSize;
	    int codeMask = maxCode - 1;
	    int freeCode = clearCode + 2;
	    int code = 0;
	    int oldCode = 0;;
	    byte prevChar = 0;

	    // Temproray storage for decompression
	    short prefix[] = new short[ 4096];
	    byte suffix[] = new byte[ 4096];
	    byte outCode[] = new byte[ 1025];
	    
	    // Variables used for writing pixels
	    int x = width;
	    int y = 0;
	    int off = 0;
	    int pass = 0;

	    // Read codes until the eofCode is encountered
	    while (true) {
		if (accumbits < codeSize) {
		    // fill the buffer if needed
		    remain -= 2;
		    while ( remain < 0 && !blockEnd) {
			block[0] = block[byteoff];
			byteoff = 0;
			// read the next block length
			int blockLength = readBuff();
			if (blockLength < 0) {
			    return;
			}
			if (blockLength == 0) {
			    blockEnd = true;
			}
			if ( readBuff( block, remain + 2, blockLength) == -1) {
			    return;
			}
			remain += blockLength;
		    }
		    // 2 bytes at a time saves checking for 
		    // accumbits < codeSize.
		    // We know we'll get enough and also that we can't 
		    // overflow since codeSize <= 12.
		    accumdata += (block[ byteoff++] & 0xFF) << accumbits;
		    accumbits += 8;
		    accumdata += (block[ byteoff++] & 0xFF) << accumbits;
		    accumbits += 8;
		}

		// Compute the code
		code = accumdata & codeMask;
		accumdata >>= codeSize;
		accumbits -= codeSize;
				
		//
		// Interpret the code
		//
		if (code == clearCode) {
		    // Clear code sets everything back to its initial value, 
		    // then reads the immediately subsequent code as 
		    // uncompressed data.

		    // Note that freeCode is one less than it is supposed 
		    // to be, this is because it will be incremented next 
		    // time round the loop

		    freeCode = clearCode + 1;
		    codeSize = initCodeSize + 1;
		    maxCode = 1 << codeSize;
		    codeMask = maxCode - 1;

		    // Continue if we've NOT reached the end, some Gif 
		    // images contain bogus codes after the last clear code.
		    if (off < raster.length) {
			continue;
		    }
		    // pretend we've reached the end of the data
		    code = eofCode;
		}

		if (code == eofCode) {
		    // make sure we read the whole block of pixels.
		    if ( !blockEnd) {
			readBuff();
		    }
		    return;
		} 

		// It must be data: save code in CurCode
		int curCode = code;
		int outCount = outCode.length;

		// If greater or equal to freeCode, not in the hash table
		// yet; repeat the last character decoded
		if (curCode >= freeCode) {
		    curCode = oldCode;
		    outCode[ --outCount] = prevChar;
		}

		// Unless this code is raw data, pursue the chain pointed
		// to by curCode through the hash table to its end; each
		// code in the chain puts its associated output code on
		// the output queue.
		while (curCode > bitMask) {
		    outCode[ --outCount] = suffix[ curCode];
		    curCode = prefix[ curCode];
		}

		// The last code in the chain is treated as raw data.
		prevChar = (byte)curCode;
		outCode[ --outCount] = prevChar;

		// Now we put the data out to the Output routine. It's
		// been stacked LIFO, so deal with it that way...
		int len = outCode.length - outCount;
		if ( len > 2 && len < x) {
		    x -= len;
		    System.arraycopy(outCode, outCount, ras, off, len);
		    off += len;
		} else while (--len >= 0) {
		    ras[ off++] = outCode[outCount++];

		    // Update the X-coordinate, and if it overflows, 
		    // update the Y-coordinate
		    if ( --x == 0) {
			// If a non-interlaced picture, just increment y 
			// to the next scan line.  If it's interlaced, 
			// deal with the interlace as described in the 
			// GIF spec.  Put the decoded scan line out
			// to the screen if we haven't gone past the bottom 
			// of it
			x = width;
			if ( interlace) {
			    switch ( pass) {
				case 0:
				    y += 8;
				    if ( y >= height) {
					pass++;
					y = 4;
				    }
				    break;
				case 1:
				    y += 8;
				    if (y >= height) {
					pass++;
					y = 2;
				    }
				    break;
				case 2:
				    y += 4;
				    if (y >= height) {
					pass++;
					y = 1;
				    }
				    break;
				case 3:
				    y += 2;
				    break;
			    }
			    off = y * width;
			}
			// Some files overrun the end
			if ( off >= ras.length) {
			    break;
			}
		    }
		} 

		// Build the hash table on-the-fly. No table is stored in
		// the file.
		prefix[freeCode] = (short)oldCode;
		suffix[freeCode] = prevChar;
		oldCode = code;

		// Point to the next slot in the table.  If we exceed the
		// maxCode, increment the code size unless
		// it's already 12.  If it is, do nothing: the next code
		// decompressed better be CLEAR
		if (++freeCode >= maxCode) {
		    if (codeSize < 12) {
			codeSize++;
			maxCode <<= 1;
			codeMask = maxCode - 1;
		    }
		}
	    }
	}

	/**
	 * Read a byte from memory, return it.
	 */
	private int readBuff() {

	    if ( buff.length < 1 + bPtr) {
		return( -1);
	    } else {
		return( ( buff[ bPtr++] & 0xFF));
	    }
	}
	
	/**
	 * Read bytes from memory, copy it.
	 * Returns number of bytes copied.
	 */
	private int readBuff( byte dst[], int off, int len) {

	    bPtr += off;
	    return( readBuff( dst, len));
	}

	/**
	 * Read bytes from memory, copy it.
	 * Returns number of bytes copied.
	 */
	private int readBuff( byte dst[], int len) {

	    if ( buff.length < len + bPtr) {
		return( -1);
	    } else {
		System.arraycopy( buff, bPtr, dst, 0, len);
		bPtr += len;
	        return( len);
	    }
	}
}

