#include "Grid.h"
#include <stdio.h>

//*****************************************************************
//* Constructor
//*****************************************************************
Grid::Grid() : cellSize(0), numCellsWide(0), startCellY(0), endCellY(0), 
numPts(0), numShapePts(0), param(), t(NULL), cells(NULL),
tPlusOneHalf(NULL), curNumShapePts(0), tMinusOneHalf(NULL), 
shapePtsT(NULL),shapePtsTMinusOneHalf(NULL),shapePtsTPlusOneHalf(NULL),
endOfPts(0), numCellsHigh(0), curShapes(0), numShapes(0),
theShapes(NULL)
{
}

//*****************************************************************
//* Constructor
//*****************************************************************
Grid::Grid(Parameter P) : cellSize(0), 
numCellsWide(0), numPts(0), numShapePts(0), 
param(P), cells(NULL), t(NULL), tPlusOneHalf(NULL), tMinusOneHalf(NULL), 
curNumShapePts(0), endOfPts(0), startCellY(0),
shapePtsT(NULL),shapePtsTPlusOneHalf(NULL),shapePtsTMinusOneHalf(NULL),
endCellY(0), numCellsHigh(0), curShapes(0), numShapes(0),
theShapes(NULL)
{
}
 
//******************************************************************
//  overloaded ostream operator (standard output)
//******************************************************************
ostream & operator << ( ostream & str, const Grid & G)
{
	for(int i = 0; i < G.endOfPts; i++)
	{
		// only output points that actually really exist
		if (!G.t[i].isEmpty())
			str << G.t[i] << endl;
	}	

	str << "\n";

	for(int j = 0; j < G.numShapePts; j++)
		str << G.shapePtsT[j] << endl;


	return str;
}	


//******************************************************************
// write to file with point IDs
//******************************************************************
void Grid::writeID(ofstream & fout)
{
	for(int i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
			fout << setprecision(15) << t[i].ID << " " 
			     << t[i].getPosition().getX() 
				<< " " << t[i].getPosition().getY()
				<< " " << t[i].getVelocity().getX()
				<< " " << t[i].getVelocity().getY()
				<< " " << t[i].getPressure()
				<< " " << t[i].getDensity()
				<< " " << t[i].getChangeDensity()
				<< " " << t[i].getChangeVelocityX()
				<< " " << t[i].getChangeVelocityY()
			        << endl;	
	}

	fout << "\n";
}	

//******************************************************************
// write to file with neighbor count
//******************************************************************
void Grid::writeNeighbor(ofstream & fout)
{
	for(int i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
			fout << t[i].ID << " " 
			     << t[i].getPosition().getX() 
				<< " " << t[i].getPosition().getY()
				<< " " << t[i].getNumNeighbors() 
			        << endl;	
	}

	fout << "\n";
}	

//******************************************************************
//******************************************************************
void Grid::writeVerbose(ofstream & fout)
{
	for(int i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
			fout << t[i].ID << " " << t[i] << endl;	
	}

	fout << "\n";

	for(int j = 0; j < numShapePts; j++)
		fout << shapePtsT[j].ID << " " << shapePtsT[j] << endl;

}	


//******************************************************************
// overloaded ofstream operator (file output)
//******************************************************************
ofstream & operator << ( ofstream & fStr, const Grid & G)
{
	for(int i = 0; i < G.endOfPts; i++)
	{
		// only output points that actually really exist
		if (!G.t[i].isEmpty())
			fStr << G.t[i];
	}	
	fStr << "\n";

	for(int j = 0; j < G.numShapePts; j++)
		fStr << G.shapePtsT[j];

	return fStr;
}	
 
 
//*****************************************************************
//* Destructor
//*****************************************************************
Grid::~Grid()
{
	int i;

        if(cells != NULL)
        {
                for(i = 0 ; i < numCellsWide; i++)
                {
                        if (cells[i] != NULL)
                        {
                                delete [] cells[i];
                        }
                }
                delete cells;
        }

        if(t != NULL)                     delete [] t;
        if(tPlusOneHalf != NULL)          delete [] tPlusOneHalf;
        if(tMinusOneHalf != NULL)         delete [] tMinusOneHalf;
        if(shapePtsT != NULL)             delete [] shapePtsT;
        if(shapePtsTPlusOneHalf != NULL)  delete [] shapePtsTPlusOneHalf;
        if(shapePtsTMinusOneHalf != NULL) delete [] shapePtsTMinusOneHalf;
        if(theShapes != NULL)             delete [] theShapes;
}	



//**********************************************************
//* This method initializes the number of points we should expect
//* to get in our Grid. We then allocate the appropriate amount
//* of memory.
//********************************************************* 
void Grid::setNumPts(int np)
{
	numPts = np; 
    if(t != NULL)             delete [] t;
	if(tPlusOneHalf != NULL)  delete [] tPlusOneHalf;
	if(tMinusOneHalf != NULL) delete [] tMinusOneHalf;

	t = new SmoothParticle[numPts];
	assert(t != NULL);
	tPlusOneHalf = new SmoothParticle[numPts];
	assert(tPlusOneHalf != NULL);
	tMinusOneHalf = new SmoothParticle[numPts];
	assert(tMinusOneHalf != NULL);
}	

//**********************************************************
//* This method allocates the shapePts array 
//********************************************************* 
void Grid::setNumShapePts(int nsp)
{
	numShapePts = nsp; 
    if(shapePtsT != NULL)             delete [] shapePtsT;
    if(shapePtsTPlusOneHalf != NULL)  delete [] shapePtsTPlusOneHalf; 
	if(shapePtsTMinusOneHalf != NULL) delete [] shapePtsTMinusOneHalf;

 	shapePtsT = new SmoothParticle[numShapePts];
	assert(shapePtsT != NULL);
	
	shapePtsTMinusOneHalf = new SmoothParticle[numShapePts];
	assert(shapePtsTMinusOneHalf != NULL);
	
	shapePtsTPlusOneHalf = new SmoothParticle[numShapePts];
	assert(shapePtsTPlusOneHalf != NULL);
	
}


//**********************************************************
//* This method allocates the 'theShapes' array 
//********************************************************* 
void Grid::setNumShapes(int ns)
{	
	curShapes = 0;
	numShapes = ns;
    if(theShapes != NULL) delete [] theShapes;
	theShapes = new Shape *[numShapes];
	assert(theShapes != NULL);
}



//**********************************************************************
// This method just reduces the number of shape pts that the grid
// thinks it has, our slave grids will have a big allocation, but
// will only use part of the array
//**********************************************************************
void Grid::updateNumShapePts(int nsp)
{
	numShapePts = nsp;
}


//**********************************************************************
// We are told the number of cells high we are resposnsible for.
//**********************************************************************
void Grid::setNumCellsHigh(int high)
{
	numCellsHigh = high;
}



//**********************************************************************
// We are told the number of cells wide we are resposnsible for.
//**********************************************************************
void Grid::setNumCellsWide(int wide)
{
	numCellsWide = wide;
}




//*********************************************************************
//* This method takes the x position of a point and returns an
//* integer representing the cell that this point belongs to.
//*********************************************************************
int Grid::hashI(double x)
{
	return (int) floor(x / cellSize);
}




//*********************************************************************
//* This method takes the y position of a point and returns an
//* integer representing the cell that this point belongs to.
//*********************************************************************
int Grid::hashJ(double y)
{
	return (int) floor(y / cellSize);
}




//**********************************************************************
// This is used by the master to set its grid size
//**********************************************************************
void Grid::setRangeMaster()
{
	startCellY = 0;
	endCellY = numCellsHigh - 1;
}


//**********************************************************************
// Initialize the size of the grid, uses the h value in the parameter 
// object
//**********************************************************************
void Grid::calcCellSize()
{
	cellSize = 2.0 * param.h;
}


//**********************************************************************
// We calculate the total number of cells high and wide including
// buffer cells.
//**********************************************************************
void Grid::calcNumCellsWideHigh(double conWidth, double conHeight)
{
	if (cellSize != 0.0)
	{
//		numCellsWide = (int)(conWidth / cellSize) + 1;	
		numCellsWide = (int)(conWidth / cellSize);	
		numCellsHigh = (int)(conHeight / cellSize) + 1;
	}
	else
	{
		numCellsWide = numCellsHigh = 0;
		cout << "Trying to calculate number of cells when cell size"
		     << " hasn't been calculated yet.\n";
	}
}



//*******************************************************************
//* This method builds the grid by placing all free and boundary 
//* particles in their appropriate cell. 
//* It assumes that 'cellSize' is known as well as 'numPts' and 
//* 'numShapePts'.  
//********************************************************************
void Grid::buildGrid()
{
	int i;

	cells = new Cell *[numCellsWide];
	assert(cells != NULL);

	for(i = 0; i < numCellsWide; i++)
	{
		cells[i] = new Cell[numCellsHigh];
		assert(cells[i] != NULL);
	}	

	int hash_i, hash_j;
	
	// build the free particle lists for each cell.
	for(i = 0; i < endOfPts; i++)
	{
		hash_i = hashI(t[i].getPositionX());
		hash_j = hashJ(t[i].getPositionY());

		if ((hash_j < startCellY) || (hash_j > endCellY)	
			|| (hash_i < 0) || (hash_i >= numCellsWide))
		{
			printf("%d  Panic!  Our grid was given a point (%d, %d) 
			       that doesn't belong to our cells!\n", debugTimeSteps,
				   hash_i, hash_j);

			// maybe we should delete the point from array???
		}
		else
			cells[hash_i][hash_j].pushFree(i);
	}

	// build the boundary particle lists for each cell.
	for(i = 0; i < numShapePts; i++)
	{
		hash_i = hashI(shapePtsT[i].getPositionX());
		hash_j = hashJ(shapePtsT[i].getPositionY());

		if ((hash_j < startCellY) || (hash_j > endCellY)	
			|| (hash_i < 0) || (hash_i >= numCellsWide))
		{
			printf("%d  Panic!  Our grid was given a point (%d, %d) 
			       that doesn't belong to our cells!\n", debugTimeSteps,
				   hash_i, hash_j);

			// maybe we should delete the point from array???
		}
		else
			cells[hash_i][hash_j].pushBnd(i);
	}

}



//*********************************************************************
//* Allocate memory for, and initialize each free particle in the grid.
//* All initial data associated with Smooth Particles must be sent in.
//*********************************************************************
void Grid::addShape(double attribute[])
{

	// We assume that the first item is the type of shape. 
	Point p(attribute[1],attribute[2]);

	int type = (int)attribute[0];

	if(curShapes < numShapes)
	{
		switch (type)
		{
			case BOX:
				theShapes[curShapes++] = 
				            new Box(p, attribute[3], attribute[4]);
				break;
			case CYLINDER:
				theShapes[curShapes++] = 
				            new Cylinder(p, attribute[3]);
				break;
			default:
				cout << "Error: Type of shape was not recognized.\n";
			break;
		}
	}
	else
		cout << "Error: Not enough room declared initially.\n";	
}
	

//*******************************************************************
// Puts the attributes of shape given by 'index' into array.
//*******************************************************************
void Grid::getShapeAttributes(int index, double attr[])
{
	//we need to collect the appropriate attribues from this shape.
	//We do this by asking the shape what type it is, and then 
	// acting accordingly.
	int type;
	Point p;
	
	// We need the following pointers, cause each shape has 
	// attributes specific to IT and not found in a general shape.
	// Thus we will use a dynamic_cast to get at `em.
	Cylinder * realCylinder = 0; // as opposed to virtual.
	Box * realBox = 0;
	
	if(0 <= index && index < curShapes)
	{
		type = theShapes[index]->getType();

		switch (type)
		{
			case BOX:
				attr[0] = type;
				realBox = 
				    dynamic_cast< Box *>(theShapes[index]);
				p = realBox->getCornerPoint();
				attr[1] = p.getX();
				attr[2] = p.getY();
				attr[3] = realBox->getWidth();
				attr[4] = realBox->getHeight();
				break;
			case CYLINDER:
				attr[0] = type;
				realCylinder = 
				    dynamic_cast< Cylinder *>(theShapes[index]);
				p = realCylinder->getCenter();
				attr[1] = p.getX();
				attr[2] = p.getY();
				attr[3] = realCylinder->getRadius();
				attr[4] = -1;
				break;
			default:
				cout << "Error: Type doesn't match\n";
		}			
	}
	else
		cout <<"Error: tried to get attributes from a shape "
		     << "that doesn't exist (yet).\n";
}	




//*********************************************************************
//* Allocate memory for, and initialize each free particle in the grid.
//* All initial data associated with Smooth Particles must be sent in.
//*********************************************************************
void Grid::addNewFreeParticle(Point pos, int newID)
{
	int index;
	int hash_i, hash_j;

	hash_i = hashI(pos.getX());
	hash_j = hashJ(pos.getY());

	if ((hash_i < numCellsWide) && (hash_i >= 0) &&
	    (hash_j < numCellsHigh) && (hash_j >= 0))
	{
		// find a spot in our array to put this new free particle
		index = getFreeIndex();

		t[index].setPosition(pos);
		t[index].setVelocity(param.initVel);
		t[index].setMass(param.initMass);
		t[index].setDensity(param.initDensity);
		t[index].setPressure(param.initPressure);
		t[index].setSmoothingLength(param.h);
		t[index].setSpeedSound(param.c);
		t[index].undeleteParticle();
		t[index].ID = newID;
		t[index].setShapeID(-1);

		tPlusOneHalf[index] = t[index];
		tMinusOneHalf[index] = t[index];
	} else {
		cout << "Panic!  Tried to add a free particle outside of Grid" 
		<< " --- hashI = " << hash_i << " and hashJ= " << hash_j << endl;
	} 
		
}			




//**********************************************************************
// add the passed SmoothParticle to our grid
// takes the three particles current, old, and new, each representing
// the particle at a different time in its evolution. 
//**********************************************************************
void Grid::addNewMigrantParticle(SmoothParticle sp_cur, 
          SmoothParticle sp_new, SmoothParticle sp_old)
{
	int index;

	// find a spot in our array to put this new free particle
	index = getFreeIndex();

	// add this particle (as long as there was space)
	if (index!=-1)
	{
		// we have memory allocated for this smooth particle, we just
		// want to copy its values over into our array
		t[index] = sp_cur;
		t[index].undeleteParticle();

		tPlusOneHalf[index] = sp_new;
		tPlusOneHalf[index].undeleteParticle();

		tMinusOneHalf[index] = sp_old;
		tMinusOneHalf[index].undeleteParticle();

		// now add this new particle to the cell LL

		int hash_i, hash_j;

		hash_i = hashI(t[index].getPositionX());
		hash_j = hashJ(t[index].getPositionY());

		if ((hash_j < startCellY) || (hash_j > endCellY)	
			|| (hash_i < 0) || (hash_i >= numCellsWide))
		{
			printf("%d  Panic!  Our fine grid was given a point 
			that doesn't belong to our cells!\n", debugTimeSteps);

			cout << "x = " << t[index].getPositionX() << endl;
			cout << "y = " << t[index].getPositionY() << endl;
			cout << "hash_j " << hash_j << endl;
			cout << "hash_i " << hash_i << endl;
			cout << "startCellY " << startCellY << endl;
			cout << "endCellY " << endCellY << endl;

			deletePt(index);

			// maybe we should delete the point from array???
		}
		else
			cells[hash_i][hash_j].pushFree(index);
	}
}			




//*********************************************************************
//* Allocate memory for, and initialize each free particle in the grid.
//* All initial data associated with Smooth Particles must be sent in.
//*********************************************************************
void Grid::addNewBndParticle(int index, Point pos, int globalID, 
                                                   int shapeID)
{
	int hash_i, hash_j;

	hash_i = hashI(pos.getX());
	hash_j = hashJ(pos.getY());

	if ((hash_i < numCellsWide) && (hash_i >= 0) &&
	    (hash_j < numCellsHigh) && (hash_j >= 0))
	{
		Point Vel(0, 0);

		shapePtsT[index].setPosition(pos);
		shapePtsT[index].setVelocity(Vel);
		shapePtsT[index].setMass(param.initMass);
		shapePtsT[index].setDensity(param.initDensity);
		shapePtsT[index].setPressure(param.initPressure);
		shapePtsT[index].setSmoothingLength(param.h);
		shapePtsT[index].setSpeedSound(param.c);
		shapePtsT[index].undeleteParticle();
		shapePtsT[index].ID = globalID;
		shapePtsT[index].setShapeID(shapeID);
		
		shapePtsTMinusOneHalf[index] = shapePtsT[index];
		shapePtsTPlusOneHalf[index]  = shapePtsT[index];
	}
	else
			cout << "Panic!  Tried to added shape particle outside of "
			     << "cell range." << endl;
}
	


//******************************************************************
//* This method determines what the error was, prints out a message
//* and exits the program. This method should be modified later to 
//* take into account errors that can be fixed dynamically, so that
//* a warning message and continuation is what happens.
//*****************************************************************
void Grid::error(char * errorMsg)
{
	cout << "An error occured in Grid for "<< errorMsg << "!\n";
	cout << "Exiting the program now.\n";
	exit(-1);
} 




//**********************************************************************
// delete the specified point (actually set its mass to zero)
// and add the position to our freeSpace list
//**********************************************************************
void Grid::deletePt(int i)
{
	Point p(0.0, 0.0);

	t[i].deleteParticle();
	t[i].setMass(0.0);
	t[i].setPositionXY(0.0, 0.0);
	t[i].setVelocityXY(0.0, 0.0);
	t[i].setDensity(0.0);
	t[i].setPressure(0.0);
	t[i].setChangeVelocity(p);
	t[i].setChangeDensity(0.0);

	tPlusOneHalf[i] = t[i];
	tMinusOneHalf[i] = t[i];
	
	freeSpace.push(i);

}



//**********************************************************************
// Returns the 'current' smoothparticle indexed at 'j'
//**********************************************************************
SmoothParticle Grid::getParticle(int j)
{
	return t[j];
}

//**********************************************************************
// return the particle with matching ID
//**********************************************************************
SmoothParticle Grid::getParticleID(int id)
{
	for (int i=0; i<getNumPts(); i++)
	{
		if (t[i].ID == id)
			return t[i];
	}

	// add so something reasonable comes back when none found
	SmoothParticle sp;
	sp.ID = -1;
	return (sp);
}

//**********************************************************************
// Returns the shapePt indexed at 'j'
//**********************************************************************
SmoothParticle Grid::getShapeParticle(int j)
{
	return shapePtsT[j];
}

//**********************************************************************
// Returns the 'new' smoothparticle indexed at 'j'
//**********************************************************************
SmoothParticle Grid::getPlusHalfParticle(int j)
{
	return tPlusOneHalf[j];
}

//**********************************************************************
// Returns the 'old' smoothparticle indexed at 'j'
//**********************************************************************
SmoothParticle Grid::getMinusHalfParticle(int j)
{
	return tMinusOneHalf[j];
}

//**********************************************************************
// output a verbose list of all the particles in the grid
//**********************************************************************
void Grid::printVerbose()
{
	cout << debugTimeSteps << " printVerbose()" << endl;
        for (int i=0; i<getNumPts(); i++)
        {
                cout << "particle " << i << ": ";
                cout << "(hash " << hashJ(
                        getParticle(i).getPositionY()) << ") ";
                cout << getParticle(i);
                cout << endl;
        }
	cout << "-----------------------------" << endl;
        for (int i=0; i<getNumShapePts(); i++)
        {
                cout << "shape particle " << i << ": ";
                cout << getShapeParticle(i);
                cout << endl;
        }               

	cout << endl << "Free space list:" << endl;

	freeSpace.print();
}

//**********************************************************************
// deletes all the free particles in the grid, used by the master
// before it receives a complete update from the slave nodes
//**********************************************************************
void Grid::deleteAllFree()
{
	int cellIndexI, cellIndexJ;
	int size, i, loop;

	for(cellIndexJ = startCellY; cellIndexJ <= endCellY; cellIndexJ++)
		for(cellIndexI = 0; cellIndexI < numCellsWide; cellIndexI++)
		{
			//Point to first free particle in this cell.
			cells[cellIndexI][cellIndexJ].cur2FirstFreeItem();
			
			//Get the number of free particles in this cell and then loop
			//over all of them, moving them if necessary.
			size = cells[cellIndexI][cellIndexJ].getNumFreePts();

			for(loop = 0; loop < size; loop++)
			{
				i = cells[cellIndexI][cellIndexJ].getCurrentFreeItem();
				
				deletePt(i);
				cells[cellIndexI][cellIndexJ].cur2NextFreeItem();
			}

			// now free up the linked list associated with this cell
			cells[cellIndexI][cellIndexJ].deleteFree();
		}

}

//**********************************************************************
// return a index of a free spot in our particle arrays
//**********************************************************************
int Grid::getFreeIndex()
{
	int index;

	if (freeSpace.isEmpty())
	{
		// our free space list is empty, so we want to add
		// this particle to the end of our array

		index = endOfPts;
	
		endOfPts++;
		if (endOfPts > numPts)
		{
			printf("Panic!  We've run out of space in our particle 
			        array.\n");

			printVerbose();
		}
	}
	else
	{

		// put this particle in an empty hole in our array
		// and delete the entry from the free space list

		freeSpace.firstCur();
		index = freeSpace.getCur();
		freeSpace.deleteCur();

		if ((index<0) || (index>=numPts))
		{
			printf("Panic!  Space obtained from freeSpace is invalid! 
			        index=%d\n", index);
			freeSpace.print();
		}
	}

	return index;
}

//**********************************************************************
// see if we can move the end of particle list pointer to the left
//**********************************************************************
void Grid::updateEndOfPts()
{
        while (t[endOfPts].isEmpty())
                endOfPts--;
        endOfPts++;
}


//**********************************************************************
// Get a packet of all the information it will need to know about this
// frame in the simulation (particle data, shape data, min/max/avg)
//
// This is used to send the client updates, also can be used to write
// files locally that can later be post processed by the client
//
// Return the number of floats it has put in the passed buffer address
//**********************************************************************

// added kdv 5/29, moved to Grid 5/31

int Grid::getFrame(float **buffer)
{
	// before we can allocate a communication buffer, we
	// must find out how big it will be.  

	int total_free = 0;
	int total_bound = 0;
	long pos = 0;
	int num_shapes = 0;
	int i, j;
	SmoothParticle sp;
	float sp_pack[SMOOTH_SIZE_CLIENT];

	// init variables to hold our calculated min, max and avgs

	double min[SMOOTH_SIZE_CLIENT];
	double max[SMOOTH_SIZE_CLIENT];
	float avg[SMOOTH_SIZE_CLIENT];

	for (i = 0; i < SMOOTH_SIZE_CLIENT; i++)
	{
		min[i] = 999999.0;
		max[i] = -999999.0;
		avg[i] = 0.0;
	}

	// Add up the free particles.
	// Also find out the min, max and average of all the attributes
	// we are sending over.

	for (i = 0; i < getEndOfPts(); i++)
	{
		sp = getParticle(i);

		if (!(sp.isEmpty()))
		{
			sp.getPacketClient(sp_pack);

			for (j = 0; j < SMOOTH_SIZE_CLIENT; j++)
			{
				if (sp_pack[j] < min[j])
					min[j] = sp_pack[j];
				if (sp_pack[j] > max[j])
					max[j] = sp_pack[j];
				avg[j] += sp_pack[j];
			}

			total_free++;
		}
	}

	// finish the average calculation

	for (j = 0; j < SMOOTH_SIZE_CLIENT; j++)
		avg[j] = avg[j] / total_free;

	total_bound = getNumShapePts();
	num_shapes = getNumShapes();

	// now we need to figure out how big our total message will be

	int message_size = 0;

	message_size += (total_free + total_bound) * SMOOTH_SIZE_CLIENT;  // for the particle data
	message_size += SMOOTH_SIZE_CLIENT * 3;				  // for the min, max, avg
	message_size += 5;						  // for number free, number bnd, container width, container height, and num shapes			
	message_size += num_shapes * MAX_ATTRS;		      		  // to send the shape info over

	*buffer = (float *) malloc (sizeof(float) * message_size);

	/*  first put in: number of free particles
			  number of bnd particles
			  container width
			  container height
			  number of shape objects    */

	(*buffer)[pos] = (float) total_free;
	pos++;
	(*buffer)[pos] = (float) total_bound;
	pos++;
	(*buffer)[pos] = (float) getContainerWidth();
	pos++;
	(*buffer)[pos] = (float) getContainerHeight();
	pos++;
	(*buffer)[pos] = (float) num_shapes;
	pos++;

	// put the min, max, avg info in there

	for (j = 0; j < SMOOTH_SIZE_CLIENT; j++)
	{
		(*buffer)[pos] = (float) min[j];
		pos++;		
		(*buffer)[pos] = (float) max[j];
		pos++;		
		(*buffer)[pos] = (float) avg[j];
		pos++;		
	}	

	// put the shape attributes in the pot
	for (j = 0; j < num_shapes; j++)
	{
		double attr[MAX_ATTRS];
		getShapeAttributes(j, attr);
	
		for (int k = 0; k < MAX_ATTRS; k++)
		{
			(*buffer)[pos] = attr[k];
			pos++;
		}	
	}

	// put the free particle into the packet

	for (i = 0; i < getEndOfPts(); i++)
	{
		sp = getParticle(i);

		if (!sp.isEmpty())
		{
			sp.getPacketClient(sp_pack);
			for (j = 0; j < SMOOTH_SIZE_CLIENT; j++)
			{
				(*buffer)[pos] = sp_pack[j];
				pos++;
			}
		}
	}

	// now put the bound particle into the packet

	for (i = 0; i < total_bound; i++)
	{
		sp = getShapeParticle(i);
		sp.getPacketClient(sp_pack);
		for (j = 0; j < SMOOTH_SIZE_CLIENT; j++)
		{
			(*buffer)[pos] = sp_pack[j];
			pos++;
		}
	}

	if (pos != message_size)
		cout << "Panic!  Something went wrong building the frame packet  pos=" << pos << " message_length=" << message_size << endl;

	return pos;
}

//**********************************************************************
// Find the particle with the biggest density difference
//**********************************************************************
void Grid::printMinMaxDensity()
{
	double max = -999999.0;
	SmoothParticle sp;
	double diff;

	for (int i = 0; i < getEndOfPts(); i++)
	{
		sp = getParticle(i);

		if (!sp.isEmpty())
		{
			diff = (sp.getMaxDensity() - sp.getMinDensity()) /
					param.initDensity;

			if (diff > max)
				max = diff;
		}
	}

	cout << "Biggest density difference was " << max*100 << "%" << endl;

	if (max*100 > 3.0)
		cout << "Oh oh!  The density difference is too big!" << endl;

}

