
// Contains an array of ints representing a bitmap image

// It also knows the size of the world coordinates, so it can map
// to screen coordinates

public class PixelFrame
{
    private int width;
    private int height;
    private float max_x;
    private float max_y;
    
    private int pixels[];
    
    public PixelFrame(int w, int h, float mx, float my)
    {
        width = w;
        height = h;
        max_x = mx;
        max_y = my;
        
        pixels = new int[width * height];
    }
    
    // clear the pixel array out to be all black
    public void clear()
    {
        for (int i = 0; i < (width * height); i++)            
             pixels[i] = (255 << 24);
    }

    // set the color at a specific pixel
    // take the (x,y) world coordinate and the (r,g,b) value
    
    public void add(float x, float y, int r, int g, int b)
    {
        int pix_y, pix_x;

        // make sure the requeted point is in our range
        if ((x < max_x) && (y < max_y) &&
            (x > 0.0) && (y > 0.0))
        {
            pix_x = hashI(x);
            pix_y = hashJ(y);

            pixels[pix_y * width + pix_x] = (255 << 24) | (r << 16) | (g << 8) | (b);        // free particle            
        }    
    }

    // color a specific pixel
    public void addAt(int i, int j, int r, int g, int b)
    {  
        // we need to invert the y coordinate since our lower left is
        // (0,0) and not our upper left
        
        int newj = (height-1) - j;
        
        pixels[newj * width + i] = (255 << 24) | (r << 16) | (g << 8) | (b);        // free particle            
    }

    // color a line between two world points
    
    public void line(float x1, float y1, float x2, float y2, int r, int g, int b)
    {
        // make a good guess at how many segments we'll need to go through
               
        int segments = (int) (Math.sqrt(hashI(x1-x2) * hashI(x1-x2) +
                                       hashJ(y1-y2) * hashJ(y1-y2)) * 1.5);
                           
        float delta_x = (x1 - x2) / segments;
        float delta_y = (x1 - x2) / segments;
        
        float cur_x = x1;
        float cur_y = y1;
        
        for (int i = 0; i < segments; i++)
        {
            add(cur_x, cur_y, r, g, b);
            cur_x += delta_x;
            cur_y += delta_y;
        }        
    }
    
    // get the array of integers
    public int [] getArray()
    {
        return pixels;
    }

    // convert a world coordinate to a screen coordinate
    private int hashI(float x)
    {     
        return (int) (width * x / max_x);
    }
    
    private int hashJ(float y)
    {
        // we need to invert the y coordinate since our lower left is
        // (0,0) and not our upper left

       return (int) (height - (height * y / max_y));
    }

}