// This class handles MPI communications between the slave and the master

#ifndef SLAVEMPI_HPP
#define SLAVEMPI_HPP

#include "SlaveWorld.h"
#include "msgtags.h"
#include "Point.h"
#include <stdlib.h>
#include <unistd.h>      // for gethostname
#include <stdio.h>
#include "Bench.h"
#include "Timer.h"

class SlaveMPI {

	int DEBUG;		// should we print debugging output?
	int our_rank;		// what our rank is in the MPI WORLD
	int north_rank;		// who is on our north side (-1 if nobody)
	int south_rank;		// who is on our south side (-1 if nobody)
	SlaveWorld *world;	

	public:
		SlaveMPI(int num_p, int nr, int sr, int debug);
		void runStats();			// run a performance test on ourself
		int getCommand();			// get a command from the master, return number of the command
		int initSim();				// receieve simulation init from master
		void simLoop();				// main simulation loop

	private:
		void sendTo(int dest);			// send buffer row and migrants to dest
		void recvFrom(int source);		// recv buffer row and migrants from source
		void sendMasterUpdate();		// send the master an update of all our body particles
		void commitReception();			// add particles in the northBuffer and southBuffer to our grid
		void sendMasterShutdown();		// tell the master we are done

		double *northBuffer;		// block of doubles recv'd from north neighbor
		double *southBuffer;		// block of doubles recv'd from south neighbor
		int northNumPts;		// how many particles are in the north buffer
		int southNumPts;		// how many particles are in the south buffer

		int updateTag;			// tag number to use in next master update
		int num_p_computing;		// how many computing slaves there are, needed in case some are sitting out
};

#endif
