//*******************************************************
//
// This class initializes everything for the simulation
// that is dependent on the SPH method (e.g. the Grid, 
// SPH variables, timestep, particle positions, but not
// anything having to do with setting up the parallel 
// implementation of this program)
//
//Written by: Brian Schlatter
// Modified by: Keith vertanen
// Last modified: 6/12/1999
//
//********************************************************

#ifndef WORLD_HPP
#define WORLD_HPP

#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include "Shape.h"
#include "Grid.h"
#include "Parameter.h"
#include "Cylinder.h"
#include "Box.h"
#include "ImageSGI.h"

#define MAX_SHAPES 20

class World {
	public:
		Grid * theGrid;
		Parameter param;
		
		World();
		~World();

		double getContainerWidth() const    { return containerWidth;    }
		double getContainerHeight() const   { return containerHeight;   }
		void setContainerWidth(double d)    { containerWidth  = d;   }
		void setContainerHeight(double d)   { containerHeight = d;   }

		void createWorld();
		void calcParticleSpacing();
		void calcSPHInitConditions();
		void calcInitSPHPositions();

		void writeFileDebug();
		void writeFileImage();
		void writeFileClient();
		void writeFileNeighbor();

		void incrementTime();
	
		int globalID;
		int outFileNum;

	private:
		double containerWidth, containerHeight;
};

#endif
